/*
 *  Extract -- extract files from disk images (D2F output files)
 *
 *  WAC  8/20/01    Original version
 */

#include "dskfile2.h"
#include "crc16.h"

#define NULL (0)

#define MAX_SPF 9

#define END_OF_FAT_CHAIN 0xffff
BYTE fatbuf[MAX_SPF * SECTOR_SIZE];

WORD spf;  /* the basics from the BPB */
WORD nfats;
WORD nroot;
WORD reserved_sectors;

WORD first_fat_sector;  /* pre-computed for our convenience */
WORD first_root_sector;
WORD first_data_sector;

struct hdr_format header;

#define NUM_SECTOR_BUFS 32
#define SECTOR_BUF_UNUSED 0xffff
struct
   {
   WORD sector_num;  /* what sector number is in the buffer? */
   WORD timeout;     /* timeout count for re-using sector */
   BYTE buf[SECTOR_SIZE];  /* the sector buffer itself */
   } sector_bufs[NUM_SECTOR_BUFS];
WORD warning_flag;  /* did a sector we use come from a "bad sector"? */

#define MAX_SECTORS (MAX_SPT * MAX_CYL * MAX_HEADS)  /* 18 * 80 * 2 = 2880 */
DWORD sector_offsets[MAX_SECTORS];  /* 18 * 80 * 2 = 2880 */

union
   {
   BYTE data[512];
   struct
      {
      BYTE fill1[11];
      WORD bps;
      BYTE spc;
      WORD rs;
      BYTE nfats;
      WORD nroot;
      WORD totalsectors;
      BYTE mediabyte;
      WORD spf;
      WORD spt;
      WORD nheads;
      DWORD hiddensects;
      DWORD ts_dword;
      } fine;
   } bpb;   /* the BPB read from the boot record */

#define NUM_BPBTABLE_ENTRIES 2 /* add more later */
struct default_bpb
   {
   WORD bps;
   BYTE spc;
   WORD rs;
   BYTE nfats;
   WORD nroot;
   WORD ts;
   BYTE mb;
   WORD spf;
   WORD spt;
   WORD nheads;
   } bpbtable[NUM_BPBTABLE_ENTRIES] =
   {    /* BPS      SPC RS NFATS NROOT TS(H*C*S) MB SPF SPT NHEADS */
      { SECTOR_SIZE,  1, 1,  2,   64, 1*40*8,  0xfe,  1,  8,  1 },  /* 160K */
      { SECTOR_SIZE,  1, 1,  2 , 224, 2*80*18, 0xf0,  9, 18,  2 }   /* 1.44M */
   };

#define ERASED_ENTRY 0xe5
#define ATTRIB_NONE      0x00
#define ATTRIB_READONLY  0x01
#define ATTRIB_HIDDEN    0x02
#define ATTRIB_SYSTEM    0x04
#define ATTRIB_LABEL     0x08
#define ATTRIB_DIRECTORY 0x10
#define ATTRIB_ARCHIVE   0x20
struct direntry
   {
   BYTE name[8];  /* not zero terminated */
   BYTE ext[3];  /* not zero terminated */
   BYTE attribute;
   BYTE reserved[10];  /* bytes 12-21 */
   WORD time;
   WORD date;
   WORD cluster;
   DWORD size;  /* DeSmet does not support unsigned long.  Sigh... */
   };

#define DIRENTRIES_PER_SECTOR (SECTOR_SIZE / sizeof(struct direntry))

#define MAX_EXTRACT_PATHS 32
BYTE extract_path[MAX_EXTRACT_PATHS][128];  /* path to extract from, no wildcards */
BYTE extract_wild[MAX_EXTRACT_PATHS][16];   /* name to extract, wildcards, accepted */
BYTE temp_path[128];  /* used to build actual path in extract_files() */
WORD num_extract_paths;
BYTE target_path[128];  /* path of name to extract */
BYTE dest_path[128];  /* add later: can put files in path other than current dir */
WORD extract_path_index = 0;
WORD recurse_dirs = FALSE;
WORD list_files_only = FALSE;
WORD add_imagename_to_path = TRUE;
WORD create_image_path = TRUE;
BYTE all_files[] = "????????.???"; /* generic wildcard filename */
                 /* 01234567.012 */
BYTE ignore_1[] = ".";  /* ignore the "ourselves" direntry */
BYTE ignore_2[] = "..";  /* ignore the "parent" direntry */

BYTE imagename[128];  /* name of the image file, wildcards accepted */
BYTE iname[16];  /* name of image file itself */
BYTE imagepath[128];  /* path to image file */
BYTE currentdir[128];  /* our current location on hard drive */

WORD num_images;

#define MIN(a,b)  (((a) < (b)) ? (a) : (b))

BYTE dta[64];  /* for find_first(), find_next() */
WORD dta_ofs,dta_seg;

DWORD lseek();
WORD normalize_filename();  /* forward definition */

int filein, fileout;  /* file handles, signed ints */

WORD file_found = FALSE;  /* global flag: we actually did something */

main(argc,argv)
   WORD argc;
   BYTE *argv[];
   {
   WORD i, j;

   for (i = 1; i < argc; i++)
      {
      if (argv[i][0] == '-')
         {
         switch (toupper(argv[i][1]))
            {
            case 'R':
               recurse_dirs = TRUE;
               break;
            case 'V':
            case 'L':
               list_files_only = TRUE;
               break;
            case 'D':
               add_imagename_to_path = FALSE;
               break;
            case 'P':
               create_image_path = FALSE;
               break;
            default:
               help();  /* does not return */
               break;
            }
         for (j = i; j < argc-1; j++)
            argv[j] = argv[j+1];
         argc--;
         i--;
         }
      }
   if (((!list_files_only) && (argc < 3))
          || ((list_files_only) && (argc < 2)))
      help();
   strcpy(imagename,argv[1]);
   num_images = 0;
   for (i = 0, j = 0; imagename[i]; i++)
      if (((imagename[i] == '\\') || (imagename[i] == ':'))
                 && (imagename[i+1]))
         j = i + 1;  /* find end of path */
   strcpy(imagepath,imagename);
   imagepath[j] = '\0';  /* truncate actual filename */
   for (num_extract_paths = 0, i = 2; i < MIN(argc,MAX_EXTRACT_PATHS+2);
                          i++,num_extract_paths++)
      {
      extract_path[num_extract_paths][0] = '\0';  /* init string */
      if (argv[i][0] != '\\')  /* if path does not start with a backslash */
         strcat(extract_path[num_extract_paths],"\\");  /* add it */
      strcat(extract_path[num_extract_paths],argv[i]);
      }
   for (j = 0; j < num_extract_paths; j++)
      {
      for (i = strlen(extract_path[j]); (i) && (extract_path[j][i] != '\\'); i--)
         ;  /* the for loop does it all -- trim (wildcard)filename from path */
      if (strlen(extract_path[j])-i > 11)  /* if wildcard len too big! */
         {
         if (i)
            i++;  /* skip \ when showing wildname */
         printf("Error: invalid wildcard name %s!\n",&extract_path[j][i]);
         exit(1);
         }
      strcpy(extract_wild[j],&extract_path[j][i+1]);  /* save wildname only */
      extract_path[j][i] = '\0';  /* truncate wildname from path */
      normalize_filename(extract_wild[j]);  /* put in form for later use */
      }
   if (find_first(imagename,0,iname,dta) == 0)
      do
         {
         if (check_image_file(imagepath,iname))
            num_images++;
         } while (find_next(imagename,0,iname,dta) == 0);
   if (!num_images)
      {
      printf("Error:  No image files match %s\n",imagename);
      exit(1);
      }
   if (num_images == 1)
      add_imagename_to_path = FALSE;
   if (find_first(imagename,0,iname,dta) == 0)
      do
         {
         if ((filein = init_image(imagepath,iname)) == -1)
            {
            printf("Unable to process file %s%s\n",imagepath,iname);
            continue;
            }
         if (list_files_only)
            show_image_filenames();
         else
            extract_image_files();
         close(filein);
         } while (find_next(imagename,0,iname,dta) == 0);
   if ((!list_files_only) && (!file_found))
      printf("No matching files found!\n");
   }

/* update crc */
WORD updatecrc(crc,c)
   WORD crc, c;
   {
   WORD tmp;

   tmp=crc^c;
   crc=(crc>>8)^crc16tab[tmp & 0xff];
   return(crc);
   }

int init_image(path,name)  /* starts similar to check_image_file(), but not exactly */
   BYTE *path, *name;
   {
   WORD i, j, count, crc;
   WORD h, c, s;  /* head, cyl, sector */
   struct sect_hdr sector_header;
   BYTE tempstr[128];

   strcpy(tempstr,path);
   strcat(tempstr,name);
   if ((filein = open(tempstr,0)) == -1)
      {
      printf("File %s inaccessable.\n",tempstr);
      return(-1);
      }
   if (read(filein,&header,sizeof(header)) != sizeof(header))
      {
      printf("File %s read error (really an image file?)\n",tempstr);
      close(filein);
      return(-1);
      }
   if (strcmp(header.signature,sig))
      {
      printf("File %s wrong signature.\n",tempstr);
      close(filein);
      return(-1);
      }
   if (header.version > VERSION)
      {
      printf("File %s unknown version.\n",tempstr);
      close(filein);
      return(-1);
      }
   if ((header.headcount > MAX_HEADS)
           || (header.cylcount > MAX_CYL)
           || (header.sectorcount > MAX_SPT))
      {
      printf("File %s image not a floppy.\n",tempstr);
      close(filein);
      return(-1);
      }
   for (i = 0; i < NUM_SECTOR_BUFS; i++)
      sector_bufs[i].sector_num = SECTOR_BUF_UNUSED;  /* reset buffers */
   count = 0;  /* sector count, index into sector_offsets[] */
   for (c = 0; c < header.cylcount; c++)
      {
      for (h = 0; h < header.headcount; h++)
         {
         for (s = 1; s <= header.sectorcount; s++)
            {
            sector_offsets[count] = lseek(filein,0L,1);  /* get current position */
            if (read(filein,&sector_header,sizeof(sector_header)) != sizeof(sector_header))
               {
               printf("Error reading %s.  It may be corrupt!\n",tempstr);
               close(filein);
               return(-1);
               }
            if ((sector_header.head != h) || (sector_header.cyl != c)
                      || (sector_header.sector != s))
               {
               printf("Error: %s is corrupt (out of sequence).\n",tempstr);
               close(filein);
               return(-1);
               }
            switch (sector_header.flags)
               {
               case SFLAG_NORMAL:
                  if ((h == 0) && (c == 0) && (s == 1))  /* if bootrecord */
                     {
                     if (read(filein,&bpb.data[0],SECTOR_SIZE) != SECTOR_SIZE)
                        {
                        printf("Error reading %s.  It may be corrupt.\n",tempstr);
                        close(filein);
                        return(-1);
                        }
                     crc = 0;
                     for (i = 0; i < SECTOR_SIZE; i++)
                        crc = updatecrc(crc,bpb.data[i]);
                     if (crc != sector_header.crc)
                        {
                        printf("Error: %s is corrupt (bad CRC!)\n",tempstr);
                        close(filein);
                        return(-1);
                        }
                     if ((bpb.fine.bps != SECTOR_SIZE)
                            || ((bpb.fine.spc != 1) && (bpb.fine.spc != 2))
                            || (bpb.fine.rs != 1)
                            || ((bpb.fine.nfats != 1) && (bpb.fine.nfats != 2))
                            || (bpb.fine.totalsectors != (header.cylcount*header.headcount*header.sectorcount)))
                            /* if invalid BPB */
                        {
                        for (i = 0; i < NUM_BPBTABLE_ENTRIES; i++)
                           {
                           if ((bpbtable[i].nheads == header.headcount)
                               && (bpbtable[i].spt == header.sectorcount)
                               && (bpbtable[i].ts == (header.headcount*header.sectorcount*header.cylcount)))
                              {
                              memcpy(&bpb.fine.bps,&bpbtable[i],sizeof(struct default_bpb));
                                 /* memcpy(DEST,SRC,LEN) */
                              break;
                              }
                           }
                        if (i == NUM_BPBTABLE_ENTRIES)  /* if can't find appropriate BPB */
                           {
                           printf("Error: %s has unknown BPB!\n",tempstr);
                           close(filein);
                           return(-1);
                           }
                        }
                     first_fat_sector = bpb.fine.rs;  /* reserved sectors */
                     first_root_sector = first_fat_sector + bpb.fine.nfats * bpb.fine.spf;
                     first_data_sector = first_root_sector + ((bpb.fine.nroot * 32) + SECTOR_SIZE-1) / SECTOR_SIZE;
                     if (bpb.fine.spf > MAX_SPF)
                        {
                        printf("Error: invalid FAT in %s!\n",tempstr);
                        close(filein);
                        return(-1);
                        }
                     }
                  else
                     if ((count >= first_fat_sector) && (count < first_fat_sector + bpb.fine.spf))
                        {
                        if (read(filein,&fatbuf[(count-first_fat_sector)*SECTOR_SIZE],SECTOR_SIZE) != SECTOR_SIZE)
                           {
                           printf("Error reading %s.  It may be corrupt.\n",tempstr);
                           close(filein);
                           return(-1);
                           }
                        crc = 0;
                        for (i = 0; i < SECTOR_SIZE; i++)
                           crc = updatecrc(crc,fatbuf[(count-first_fat_sector)*SECTOR_SIZE+i]);
                        if (crc != sector_header.crc)
                           {
                           printf("Error: %s is corrupt (bad CRC!)\n",tempstr);
                           close(filein);
                           return(-1);
                           }
                        }
                     else
                        lseek(filein,(DWORD)SECTOR_SIZE,1);  /* skip data for now */
                  break;
               case SFLAG_BAD:
                  if ((h == 0) && (c == 0) && (s == 1))  /* if bootrecord */
                     {
                     printf("Error:  boot record of %s not intact.\n",tempstr);
                     close(filein);
                     return(-1);
                     }
                  break;
               case SFLAG_REPEAT:
                  if ((h == 0) && (c == 0) && (s == 1))  /* if bootrecord */
                     {
                     printf("Error:  boot record of %s not intact.\n",tempstr);
                     close(filein);
                     return(-1);
                     }
                  break;
               default:
                  printf("Error: %s corrupt (bad sector marker).\n",tempstr);
                  close(filein);
                  return(-1);
                  break;
               }
            count++;  /* next sector */
            }  /* for (s) */
         } /* for (h) */
      } /* for (c) */
   return(filein);
   }

void *get_sector(sectornum)
             /* NOTE: since this is NOT a multitasking program, I don't
              *       have to worry about re-assigning a currently used
              *       sector buffer.
              */
   WORD sectornum;
   {
   WORD i, j, k, crc;
   struct sect_hdr sector_header;

   for (i = 0, j = 0, k = NUM_SECTOR_BUFS; i < NUM_SECTOR_BUFS; i++)
      {
      if (sector_bufs[i].sector_num == sectornum)
         {
         sector_bufs[i].timeout = NUM_SECTOR_BUFS;  /* max timeout */
         return((void *)sector_bufs[i].buf);
         }
      if (sector_bufs[i].timeout)
         sector_bufs[i].timeout--;   /* update timeout counters */
      if (sector_bufs[i].sector_num == SECTOR_BUF_UNUSED)
         {
         j = i;  /* we have our target */
         break;
         }
      if (sector_bufs[i].timeout < k)
         {
         k = sector_bufs[i].timeout;
         j = i;  /* for now, our target */
         }
      }
   sector_bufs[j].sector_num = sectornum;  /* flag what's in this sector */
   sector_bufs[j].timeout = NUM_SECTOR_BUFS;  /* max timeout */
   lseek(filein,sector_offsets[sectornum],0);  /* goto sector location */
   if (read(filein,&sector_header,sizeof(sector_header)) != sizeof(sector_header))
      {
      printf("Read error in %s%s!\n",imagepath,iname);
      return(NULL);
      }
   switch (sector_header.flags)
      {
      case SFLAG_NORMAL:
         if (read(filein,&sector_bufs[j].buf,SECTOR_SIZE) != SECTOR_SIZE)
            {
            printf("Error reading %s%s.  It may be corrupt.\n",imagepath,iname);
            return(NULL);
            }
         crc = 0;
         for (i = 0; i < SECTOR_SIZE; i++)
            crc = updatecrc(crc,sector_bufs[j].buf[i]);
         if (crc != sector_header.crc)
            {
            printf("Error: %s%s is corrupt (bad CRC!)\n",imagepath,iname);
            return(NULL);
            }
         break;
      case SFLAG_BAD:
         MEMFILL(sector_bufs[j].buf,SECTOR_SIZE,0);  /* clear buffer */
         warning_flag = TRUE;  /* we're using a bad sector to reconstruct a file */
         break;
      case SFLAG_REPEAT:
         MEMFILL(sector_bufs[j].buf,SECTOR_SIZE,sector_header.crc);  /* set buf */
         break;
      default:
         printf("Error: %s%s corrupt (bad sector marker).\n",imagepath,iname);
         return(NULL);
         break;
      }
   return((void *)sector_bufs[j].buf);
   }

WORD nextfat(cluster)  /* NOTE: only for 12-bit FATs */
   WORD cluster;
   {
   WORD temp, next;

   temp = cluster + (cluster >> 1);  /* cluster * 3 / 2  (12 bits/FAT) */
   next = ((WORD)(fatbuf[temp] + (fatbuf[temp+1]<<8)));  /* get it as a WORD */
   if (cluster & 1)  /* if ODD */
      next = (next >> 4);  /* want upper 12-bits */
   next &= 0x0fff;  /* isolate 12-bit value */
   if ((next < 2) || (next > 0xff6))
      next = END_OF_FAT_CHAIN;
   return(next);
   }

build_filename(dest,d)
   BYTE *dest;
   struct direntry *d;
   {
   WORD i;

   for (i = 0; (i < 8) && (d->name[i] != ' '); i++)
      *(dest++) = d->name[i];
   if (d->ext[0] != ' ')
      {
      *(dest++) = '.';
      for (i = 0; (i < 3) && (d->ext[i] != ' '); i++)
         *(dest++) = d->ext[i];
      }
   *dest = '\0';  /* set EOS for C */
   }
 
BYTE *format_date(str,date)
   BYTE *str;
   WORD date;
   {
   sprintf(str,"%02d/%02d/%02d",(date >> 5) & 0x0f, date & 0x1f, ((date >> 9) + 80) % 100);
   return(str);
   }

BYTE *format_time(str,time)
   BYTE *str;
   WORD time;
   {
   sprintf(str,"%02d:%02d%c",(time >> 11) % 12, (time >> 5) & 0x3f, (time >> 11) > 12 ? 'p' : 'a');
   return(str);
   }

show_subdir_names(path,cluster)  /* WARNING: RECURSIVE SUBROUTINE! */
   BYTE *path;
   WORD cluster;
   {
   WORD pathindex;
   WORD i, j, s, clustersave;
   struct direntry *d;
   void *p;
   BYTE fname[16];  /* target string */
   BYTE datestr[16];  /* for both date & time */

   pathindex = strlen(target_path);  /* remember where to truncate */
   strcat(target_path,path);  /* add path for display */
   strcat(target_path,"\\");
   clustersave = cluster;
   do
      {
      for (s = 0; s < bpb.fine.spc; s++)
         {
         if (!(p = get_sector((cluster-2)*bpb.fine.spc + first_data_sector + s)))
            return;
         for ((void *)d = p, j = 0; (d->name[0]) && (j < DIRENTRIES_PER_SECTOR); d++, j++)
                                    /* while we have names! */
            {
            if ((d->name[0] != ERASED_ENTRY) && (!(d->attribute & (ATTRIB_DIRECTORY|ATTRIB_LABEL))))
                     /* if not erased, not DIR, and not LABEL */
               {
               build_filename(fname,d);
               printf("%s%s",target_path,fname);
               for (i = strlen(target_path)+strlen(fname); i < 48; i++)
                  printf(" ");
               printf("%8ld %s",d->size,format_date(datestr,d->date));
               printf(" %s\n",format_time(datestr,d->time));
               }
            }
         if (j < DIRENTRIES_PER_SECTOR)
            break;   /* name[0] == '\0', at end of used dir entries */
         }
      } while ((j == 16) && ((cluster = nextfat(cluster)) != END_OF_FAT_CHAIN));
   cluster = clustersave;  /* back to the start */
   do
      {
      for (s = 0; s < bpb.fine.spc; s++)
         {
         if (!(p = get_sector((cluster-2)*bpb.fine.spc + first_data_sector + s)))
            return;
         for ((void *)d = p, j = 0; (d->name[0]) && (j < DIRENTRIES_PER_SECTOR); d++, j++)
                                    /* while we have names! */
            {
            if ((d->name[0] != ERASED_ENTRY) && (d->attribute & ATTRIB_DIRECTORY))
                     /* if not erased and is a DIR */
               {
               build_filename(fname,d);
               if ((strcmp(fname,".")) && (strcmp(fname,"..")))
                  show_subdir_names(fname,d->cluster);
               }
            }
         if (j < DIRENTRIES_PER_SECTOR)
            break;   /* name[0] == '\0', at end of used dir entries */
         }
      } while ((j == 16) && ((cluster = nextfat(cluster)) != END_OF_FAT_CHAIN));
   target_path[pathindex] = '\0';  /* remove path */
   }

show_image_filenames()
   {
   WORD i, j;
   struct direntry *d;
   void *p;
   BYTE fname[16];  /* target string */
   BYTE datestr[16];  /* for both date & time */

        /* first, the files in the root */
   for (i = first_root_sector; i < first_data_sector; i++)
      {
      if (!(p = get_sector(i)))
         return;
      for ((void *)d = p, j = 0; (d->name[0]) && (j < DIRENTRIES_PER_SECTOR); d++, j++)
                                 /* while we have names! */
         {
         if ((d->name[0] != ERASED_ENTRY) && (!(d->attribute & (ATTRIB_DIRECTORY|ATTRIB_LABEL))))
                     /* if not erased, not DIR, and not LABEL */
            {
            build_filename(fname,d);
            printf("%-48s %8ld %s",fname,d->size,format_date(datestr,d->date));
            printf(" %s\n",format_time(datestr,d->time));
            }
         }
      if (j < DIRENTRIES_PER_SECTOR)
         break;   /* name[0] == '\0', at end of used dir entries */
      }
          /* next, the subdirs off the root */
   strcpy(target_path,"\\");  /* start at the root */
   for (i = first_root_sector; i < first_data_sector; i++)
      {
      if (!(p = get_sector(i)))
         return;
      for ((void *)d = p, j = 0; (d->name[0]) && (j < DIRENTRIES_PER_SECTOR); d++, j++)
                                 /* while we have names! */
         {
         if ((d->name[0] != ERASED_ENTRY) && (d->attribute & ATTRIB_DIRECTORY))
                     /* if not erased and a DIR */
            {
            build_filename(fname,d);
            show_subdir_names(fname,d->cluster);
            }
         }
      if (j < DIRENTRIES_PER_SECTOR)
         break;   /* name[0] == '\0', at end of used dir entries */
      }
   }

int verify_path(path)  /* verify path exists, create if necessary */
   BYTE *path;
   {
   WORD i, start;
   BYTE savedchar;

   if (!path[0])  /* if null path */
      return(TRUE);

   strcpy(currentdir,"\\");  /* start at the root */
   getdir(&currentdir[1]);   /* get current dir */
   i = 0;  /* init vars */
   start = 0;
   while (TRUE)  /* use return to get out */
      {
      if ((path[i] == '\\') || (!path[i]))  /* if end of path element */
         {
         savedchar = path[i];  /* save this byte */
         path[i] = '\0';  /* set EOS for C */
         if (chdir(&path[start]))
            {
            if (mkdir(&path[start]))
               {
               printf("Error: unable to create output path %s.  Disk full?\n",&path[start]);
               path[i] = savedchar;  /* fixup string */
               chdir(currentdir);  /* go back "home" */
               return(FALSE);  /* return error */
               }
            if (chdir(&path[start]))
               {
               printf("Error: unable to access directory %s!\n",&path[start]);
               path[i] = savedchar;  /* fixup string */
               chdir(currentdir);  /* go back "home" */
               return(FALSE);  /* return error */
               }
            }  /* if (!chdir()) */
         path[i] = savedchar;  /* fixup string */
         if (savedchar)
            start = ++i;  /* set new start */
         else
            start = i + 1;
         }
      if (!path[i])
         {
         chdir(currentdir);  /* go back "home" */
         return(TRUE);  /* we're done */
         }
      i++;
      }  /* while TRUE */
   }

int create_output_file(path,d)  /* returns an int because it's a file handle */
   BYTE *path;
   struct direntry *d;
   {
   WORD i;

   temp_path[0] = '\0';  /* clear string */
   if (add_imagename_to_path)
      {
      strcpy(temp_path,iname);
      for (i = 0; temp_path[i]; i++)
         if (temp_path[i] == '.')  /* find extension */
            {
            temp_path[i] = '\0';  /* and remove it */
            break;
            }
      strcat(temp_path,"\\");
      }
   if (create_image_path)
      {
      if (path[0] == '\\')  /* if relative to root */
         strcat(temp_path,&path[1]);  /* make it relative to current dir */
      else
         strcat(temp_path,path);
      }
   if ((i = strlen(temp_path)) && (temp_path[i] == '\\'))
      temp_path[i] = '\0';  /* remove any trailing '\' */
   if (!(verify_path(temp_path)))
      {
      printf("Error:  Unable to verify/create output path %s (disk full?)\n",temp_path);
      return(-1);  /* if can't verify/create output path, oops! */
      }
   if (temp_path[0])  /* if we have a path */
      strcat(temp_path,"\\");  /* finish path */
   build_filename(&temp_path[strlen(temp_path)],d);  /* add name to path */
   if ((fileout = open(temp_path,0)) != -1)  /* if file already exists */
      {
      close(fileout);
      printf("Warning: %s already exists!\n",temp_path);
      printf("   Do you want to overwrite this file (Y/N)?");
      i = toupper(ci());  /* beware of macro implementations */
      printf("\n");
      if (i != 'Y')
         return(-1);
      }
   return(creat(temp_path));
   }

int cmp_filenames(name1, name2, where)  /* name1 can contain the '?' wildcard character */
   BYTE *name1, *name2;  /* NOTE: "raw" dir entries, NOT TERMINATED STRINGS! */
WORD where;
   {
   WORD i;

   for (i = 0; i < 11; i++)
      if ((name1[i] != '?') && (toupper(name1[i]) != toupper(name2[i])))
         return(1);  /* no match */
   return(0);  /* a match! */
   }

WORD normalize_filename(name)  /* TRUE if name contains wildcards, else FALSE */
   BYTE *name;
   {
   BYTE tempname[16];
   WORD n, t, has_wildcards;

   has_wildcards = FALSE;
   for (t = 0, n = 0; t < 8; t++)
      {
      if (name[n] == '*')
         {
         for (; t < 8; t++)
            tempname[t] = '?';  /* set wildcard character(s) */
         t--;  /* prepare for t++ at end of for() loop */
         while ((name[n]) && (name[n] != '.'))
            n++;  /* find EOS or extension */
         has_wildcards = TRUE;
         }
      else
         if (name[n] == '.')  /* at extension? */
            {
            for (; t < 8; t++)
               tempname[t] = ' ';  /* blank fill rest of name */
            t--;  /* prepare for t++ at end of for() loop */
            }
         else
            {
            if (name[n] == '?')
               has_wildcards = TRUE;
            tempname[t] = name[n++];
            }
      }  /* for (t...)  name loop */
   if (name[n] == '.')  /* if we have an extension */
      {
      n++;  /* skip period */
      for (; t < 11; t++)
         {
         if (name[n] == '*')
            {
            for (; t < 11; t++)
               tempname[t] = '?';  /* set wildcard character(s) */
            has_wildcards = TRUE;
            }
         else
            if (!name[n])  /* at end of extension? */
               {
               for (; t < 11; t++)
                  tempname[t] = ' ';  /* blank fill rest of name */
               }
            else
               {
               if (name[n] == '?')
                  has_wildcards = TRUE;
               tempname[t] = name[n++];
               }
         }  /* for (t...)  extension loop */
      }
   else
      {
      for (; t < 11; t++)
         tempname[t] = ' ';   /* add blank extension */
      }
   for (t = 0; t < 11; t++)
      name[t] = tempname[t];  /* NOTE: these are NOT terminated strings! */
   return(has_wildcards);
   }

make_extracted_file(temppath,d)
   BYTE *temppath;
   struct direntry *d;
   {
   DWORD len;
   WORD s, size, cluster;
   WORD time, date;
   BYTE fname[16];  /* target string */
   BYTE *p;

   warning_flag = FALSE;
   file_found = TRUE; /* found a file to extract! */
   build_filename(fname,d);
   time = d->time;
   date = d->date;  /* save these, our d will get clobbered */
   if ((fileout = create_output_file(temppath,d)) != -1)
      {
      printf("Extracting: %-12s (%-12s) --> %s\n",iname,fname,temp_path);
      len = d->size;  /* remaining length */
      cluster = d->cluster;  /* current cluster */
      while (len)  /* until I get all bytes */
         {
         for (s = 0; (len) && (s < bpb.fine.spc); s++)
            {
            size = (WORD)MIN(SECTOR_SIZE,len);
            if (!(p = get_sector((cluster-2)*bpb.fine.spc + first_data_sector + s)))
               {
               printf("Error:  read error in %s!\n",iname);
               close(fileout);
               fileout = -1;
               unlink(temp_path);
               len = 0;  /* abort all loops */
               }
            else
               {
               if (write(fileout,p,size) != size)
                  {
                  printf("Error writing %s (disk full?)\n",temp_path);
                  close(fileout);
                  fileout = -1;
                  unlink(temp_path);
                  len = 0;  /* abort all loops */
                  }
               else
                  len -= size;
               }
            }  /* for (s...) */
         if ((len) && ((cluster = nextfat(cluster)) == END_OF_FAT_CHAIN))
            {
            printf("Error: %s is corrupt (too few allocated clusters)!\n",fname);
            close(fileout);
            fileout = -1;
            unlink(temp_path);
            break;  /* abort loop! */
            }
         }  /* while (len) */
      if (fileout != -1)  /* if an error did not occur */
         {
         set_file_time_date(fileout,date,time);
         if (close(fileout) == -1)  /* if OOPS! */
            {
            printf("Error: buffered write to %s failed!\n",temp_path);
            unlink(temp_path);
            }
         else
            if (warning_flag)
               {
               printf("WARNING: %s contained a bad sector on the original disk!\n",fname);
               printf("         EXTRACTED FILE MAY BE UNUSABLE!\n");
               }
         }
      }  /* if (fileout...) */
   }

WORD wanted_file(d)
   struct direntry *d;
   {
   WORD i;

   for (i = 0; i < num_extract_paths; i++)
      if (cmp_filenames(extract_wild[i],d->name,2) == 0)
         return(TRUE);
   return(FALSE);
   }

WORD wanted_directory(path)
   BYTE *path;
   {
   WORD i, j;

   for (i = 0; i < num_extract_paths; i++)
      {
      for (j = 0; (extract_path[i][j]) && (path[j]); j++)
         if (toupper(extract_path[i][j]) != toupper(path[j]))
            break;
      if ((!extract_path[i][j]) && (!path[j]))  /* both at EOS */
         return(TRUE);
      if ((((!extract_path[i][j]) && (path[j] == '\\'))
              || (j == 0)) && (recurse_dirs))
                      /* if EOS of user specified path, next path element
                         on passed path, and we're recursing subdirectories */
         return(TRUE);
      }
   return(FALSE);  /* didn't find a matching path */
   }

extract_subdir_files(path,cluster)  /* WARNING: RECURSIVE SUBROUTINE! */
   BYTE *path;
   WORD cluster;
   {
   WORD pathindex;
   WORD i, j, s;
   struct direntry *d;
   struct direntry *p;
   BYTE fname[16];  /* target string */

   pathindex = strlen(target_path);  /* remember where to truncate */
   strcat(target_path,path);  /* add path for target name */
   strcat(target_path,"\\");
   do
      {
      for (s = 0; s < bpb.fine.spc; s++)
         {
         if (!(p = get_sector((cluster-2)*bpb.fine.spc + first_data_sector + s)))
            {
            printf("Error:  bad read/corrupt image file!\n");
            return;
            }
         for (d = p, j = 0; (d->name[0]) && (j < DIRENTRIES_PER_SECTOR); d++, j++)
                                    /* while we have names! */
            {
            if ((d->name[0] != ERASED_ENTRY) && (!(d->attribute & ATTRIB_LABEL)))
                     /* if not erased and not LABEL */
               {
               build_filename(fname,d);
               if (d->attribute & ATTRIB_DIRECTORY)  /* if a subdir */
                  {
                  if ((strcmp(fname,".") ) && (strcmp(fname,"..")))
                     {
                     extract_subdir_files(fname,d->cluster);
                         /* after this, we have to make sure we have our
                            buffer back!
                         */
                     if (!(p = get_sector((cluster-2)*bpb.fine.spc + first_data_sector + s)))
                        {
                        printf("Error:  bad read of image file!\n");
                        return;
                        }
                     d = &p[j];
                     }
                  }
               else
                  {
                  if ((wanted_directory(target_path)) && (wanted_file(d)))
                     {
                     make_extracted_file(target_path,d);
                      /* after this, we have to make sure we have our
                         buffer back!
                      */
                     if (!(p = get_sector((cluster-2)*bpb.fine.spc + first_data_sector + s)))
                        {
                        printf("Error:  bad read of image file!\n");
                        return;
                        }
                     d = &p[j];
                     }
                  }
               }
            }
         if (j < DIRENTRIES_PER_SECTOR)
            break;   /* name[0] == '\0', at end of used dir entries */
         }
      } while ((j == 16) && ((cluster = nextfat(cluster)) != END_OF_FAT_CHAIN));
   target_path[pathindex] = '\0';  /* remove path */
   }

extract_image_files()
   {
   WORD sectnum, j;
   struct direntry *d;
   struct direntry *p;
   BYTE fname[16];  /* target string */

        /* start tree search in the root */
   target_path[0] = '\0';  /* clear target name */
   for (sectnum = first_root_sector; sectnum < first_data_sector; sectnum++)
      {
      if (!((void *)p = get_sector(sectnum)))
         {
         printf("Error:  bad read of image file!\n");
         return;
         }
      for (d = p, j = 0; (d->name[0]) && (j < DIRENTRIES_PER_SECTOR); d++, j++)
                                 /* while we have names! */
         {
         if ((d->name[0] != ERASED_ENTRY) && (!(d->attribute & ATTRIB_LABEL)))
                     /* if not erased and not LABEL */
            {
            build_filename(fname,d);
            if (d->attribute & ATTRIB_DIRECTORY)  /* if a DIR */
               {
               extract_subdir_files(fname,d->cluster);
                      /* after this, we have to make sure we have our
                         buffer back!
                      */
               if (!(p = get_sector(sectnum)))
                  {
                  printf("Error:  bad read of image file!\n");
                  return;
                  }
               d = &p[j];
               }
            else
               {
               if ((wanted_directory(target_path)) && (wanted_file(d)))
                  {
                  make_extracted_file(target_path,d);
                      /* after this, we have to make sure we have our
                         buffer back!
                      */
                  if (!(p = get_sector(sectnum)))
                     {
                     printf("Error:  bad read of image file!\n");
                     return;
                     }
                  d = &p[j];
                  }
               }
            }
         }
      if (j < DIRENTRIES_PER_SECTOR)
         break;   /* name[0] == '\0', at end of used dir entries */
      }
   }

set_file_time_date(fileout,date,time)
   int fileout;
   WORD date, time;
   {
#asm
   mov  ax,5701h           ;fn = set file time/date
   mov  bx,word [bp+4]     ;get handle
   mov  cx,word [bp+8]     ;get time
   mov  dx,word [bp+6]     ;get date
   int  21h                ;call DOS
#endasm
   }

int check_image_file(path,name)
   BYTE *path, *name;
   {
   BYTE tempstr[128];

   strcpy(tempstr,path);
   strcat(tempstr,name);
   if ((filein = open(tempstr,0)) == -1)
      {
      printf("File %s inaccessable.\n",tempstr);
      return(FALSE);
      }
   if (read(filein,&header,sizeof(header)) != sizeof(header))
      {
      printf("File %s read error (really an image file?)\n",tempstr);
      close(filein);
      return(FALSE);
      }
   close(filein);
   if (strcmp(header.signature,sig))
      {
      printf("File %s wrong signature.\n",tempstr);
      return(FALSE);
      }
   if (header.version > VERSION)
      {
      printf("File %s unknown version.\n",tempstr);
      return(FALSE);
      }
   if ((header.headcount > MAX_HEADS)
           || (header.cylcount > MAX_CYL)
           || (header.sectorcount > MAX_SPT))
      {
      printf("File %s image not a floppy.\n",tempstr);
      return(FALSE);
      }
   return(TRUE);
   }

help()  /* does not return */
   {
   puts("The format is:\n");
   puts("   EXTRACT <wildcard IMGname> <wildcard filename1>\n");
   puts("                              <wildcard filename2> <...> [-R] [-L] [-D] [-P]\n");
   puts("\n");
   puts("wildcard IMGname:  name of disk image file(s), the output of the\n");
   puts("                   D2F program.  Wildcards are acceptable.\n");
   puts("wildcard filename(s): Name of the file(s) to extract.  May contain a\n");
   puts("                      path in the image. Wildcards are acceptable.\n");
   puts("-R  : optional flag.  Recurse the directory tree.  Files are assumed\n");
   puts("                      to be in the ROOT directory (or a specified\n");
   puts("                      directory).  Using -R will search all lower-level\n");
   puts("                      directories for the specified file(s) as well.\n");
   puts("-L  : optional flag.  List directory contents only, do not extract files.\n");
   puts("-D  : optional flag.  By default, if multiple image files, put output from each\n");
   puts("                      image file in it's own directory (named the same as the\n");
   puts("                      image file.)  The -D flag turns this off.\n");
   puts("-P  : optional flag.  By default, EXTRACT re-creates the directory structure\n");
   puts("                      in the image file.  The -P flag turns this off.\n");
   puts("\n\n\n\nPress any key to continue...");
   ci();
   cls();
   puts("Examples:\n");
   puts("   extract mydisk.img somefile.exe\n");
   puts("         extract SOMEFILE.EXE from the disk image MYDISK.IMG\n");
   puts("   extract mydisk.img \\sources\\*.c\n");
   puts("         extract all .C files from the directory \\SOURCES in MYDISK.IMG\n");
   puts("   extract *.img *.zip -r\n");
   puts("         extract all .ZIP files from all directories in all disk images.\n");
   exit(1);
   }

int find_first(wildname,attrib,fname,dta_addr)
   char *wildname, *fname, *dta_addr;
   int attrib;
   {
   int return_val;

   return_val = 0;
#asm
   mov ah,2fh            ;fn = get DTA
   int 21h               ;call DOS
   mov word dta_ofs_,bx  ;save old DTA offset
   mov word dta_seg_,es  ;save old DTA segment
   push ds               ;copy DS
   pop es                ;to ES
   mov dx,word [bp+10]   ;DS:DX = new DTA
   mov ah,1ah            ;fn = set DTA
   int 21h               ;call DOS
   mov dx,word [bp+4]    ;DS:DX = wild card fname
   mov cx,word [bp+6]    ;attribute
   mov ah,4eh            ;fn = find first
   int 21h               ;call DOS
   jae f_f_1             ;JNC f_f_1
   mov word [bp-2],ax    ;else save error
f_f_1:
   mov di,word [bp+8]    ;ES:DI = target string
   mov si,word [bp+10]   ;DS:SI = this DTA
   add si,30             ;DS:SI = source string
   cld                   ;UP!
f_f_2:
   lodsb                 ;get byte
   stosb                 ;store byte
   cmp al,0              ;EOS?
   jne f_f_2             ;no, continue
   mov dx,word dta_ofs_  ;get old DTA offset
   mov ds,word dta_seg_  ;get old DTA segment
   mov ah,1ah            ;fn = set DTA
   int 21h               ;call DOS
   push es               ;copy ES
   pop ds                ;to DS
#endasm
   return(return_val);
   }

int find_next(wildname,attrib,fname,dta_addr)
   char *wildname, *fname, *dta_addr;
   int attrib;
   {
   int return_val;

   return_val = 0;
#asm
   mov ah,2fh            ;fn = get DTA
   int 21h               ;call DOS
   mov word dta_ofs_,bx  ;save old DTA offset
   mov word dta_seg_,es  ;save old DTA segment
   push ds               ;copy DS
   pop es                ;to ES
   mov dx,word [bp+10]   ;DS:DX = new DTA
   mov ah,1ah            ;fn = set DTA
   int 21h               ;call DOS
   mov ah,4fh            ;fn = find next
   int 21h               ;call DOS
   jae f_n_1             ;JNC f_n_1
   mov word [bp-2],ax    ;else save error
f_n_1:
   mov di,word [bp+8]    ;ES:DI = target string
   mov si,word [bp+10]   ;DS:SI = this DTA
   add si,30             ;DS:SI = source string
   cld                   ;UP!
f_n_2:
   lodsb                 ;get byte
   stosb                 ;store byte
   cmp al,0              ;EOS?
   jne f_n_2             ;no, continue
   mov dx,word dta_ofs_  ;get old DTA offset
   mov ds,word dta_seg_  ;get old DTA segment
   mov ah,1ah            ;fn = set DTA
   int 21h               ;call DOS
   push es               ;copy ES
   pop ds                ;to DS
#endasm
   return(return_val);
   }

truncate(name,pos)
   BYTE *name;
   DWORD pos;
   {
#asm
   mov  dx,word [bp+4]     ;ds:dx = name
   mov  ax,3d02h           ;fn = open for read/write
   int  21h                ;call DOS
   jc   truncate_exit      ;nope, quit!
   mov  bx,ax              ;handle to bx
   mov  cx,word [bp+8]     ;cx:dx = location in file
   mov  dx,word [bp+6]
   mov  ax,4200h           ;fn = lseek BOF+CX:DX
   int  21h
   mov  dx,word [bp+4]     ;ds:dx = don't care pointer
   mov  cx,0               ;size to write = 0 (truncate file)
   mov  ah,40h             ;fn = write to file
   int  21h                ;call DOS
   mov  ah,3eh             ;fn = close file
   int  21h                ;call DOS
truncate_exit:
#endasm
   }

cls()
   {
#asm
   mov  ah,0fh            ;fn = get video mode
   int  10h               ;call BIOS
   mov  ah,0              ;fn = set video mode (al = mode from above)
   int  10h               ;call BIOS
#endasm
   }

void getdir(str)
   BYTE *str;
   {
#asm
   mov dl,0                   ;default drive
   mov si,word [bp+4]         ;get destination pointer
   mov ah,47h                 ;fn = GET DIR
   int 21h                    ;call DOS
#endasm
   }

int chdir(str)
   BYTE *str;
   {
#asm
   mov dx,word [bp+4]    ;DS:DX = pathname
   mov ah,3bh            ;fn = CHDIR
   int 21h               ;call DOS
   mov  ax,0             ;default to OK
   adc  ax,0             ;if CY, set error in AX
#endasm
   }

int mkdir(name)
   BYTE *name;
   {
#asm
   mov  dx,word [bp+4]   ;ds:dx = name
   mov  ah,39h           ;fn = MKDIR
   int  21h              ;call DOS
   mov  ax,0             ;default to OK
   adc  ax,0             ;if CY, set error in AX
#endasm
   }

memcpy(dest,src,len)  /* I'm sure there is a library function, but I'm not */
   BYTE *dest, *src;  /* going to look it up now.  This routine does not */
   WORD len;          /* handle overlapping moves gracefully. */
   {
   while (len--)
      *(dest++) = *(src++);
   }
